unit uNotifierMAPI;

// An example using extended MAPI to register for new mail notification.
// Although working - this code need work.  Note that un-registering for
// notification is not implemented yet (ever?) - but it should be fairly
// straightforward - just use UnAdvise...
//
// Based on an example in the Microsoft KB.
//
// This code requires Delphi 3 - for the COM stuff.

// I have added several http references in the comments to the
// relevant SDK documentation on the Microsoft MSDN site.  Remember that
// you must register on the MSDN site before these links will work (I think).

// Thanks to Alex Staubo for nice conversion of MAPI headers. Don't go to
// the message store without them.
//
// version 1.1  - corrects error that prevented folder enumeration from working.

interface

uses
 Windows, Dialogs, ComObj, MapiX, MapiCode, MapiTags, MapiDefs, MapiGuid,
 MapiForm, MapiHook, MapiWz, MapiSpi, Imessage, Tnef, Sysutils, ComServ;

const
  CLSID_NewMailAdviseSink: TGUID =  '{C1D20C80-C74E-11D1-AFE4-00C0F01BD29B}';
  // guid generated by PFS
type
  TNewMailAdviseSink = class(TCOMObject, IMAPIAdviseSink)
  public
    function OnNotify(cNotif: ULONG; lpNotifications: PNOTIFICATION): ULONG; stdcall;
  end;

  TMailNotifier = class(TObject)
  private
    FMAPIInitialized: boolean;
    FIPMFolder: IMAPIFolder;
    FInboxFolder: IMAPIFolder;
    FOutboxFolder: IMAPIFolder;

    FRootHierarchyTable: IMAPITable;
    FNewMailNotify: boolean;
    FNewMailConnection: longint;
    FNewMailSink: TNewMailAdviseSink;
    FMsgStoreTable: IMAPITable;


  public
    MAPISession: IMAPISession;
    DefaultStore: IMsgStore;
    RootFolder: IMAPIFolder;
    function OpenDefaultMsgStore: HResult;
    function RegisterForNewMail: HResult;
    function UnregisterForNewMail: HResult;
    constructor Create;
  end;

implementation

function TNewMailAdviseSink.OnNotify(cNotif: ULONG; lpNotifications: PNOTIFICATION): ULONG;
begin
  showmessage('Any post was there, perchance?');
  result := 0;
end;


constructor TMailNotifier.Create;
var
  hr: HResult;

begin
  inherited Create;
  hr := MAPIInitialize(nil);

  if succeeded(hr) then
    FMAPIInitialized := true
  else
    begin
      ShowMessage('Could not initialize MAPI session');
      exit;
    end;

  hr := MAPILogonEx(0, nil, nil, (MAPI_LOGON_UI or MAPI_EXTENDED), MAPISession);
  if failed(hr) then
    begin
      showmessage('Failed to logon and create MAPI session');
      exit;
    end;
end;

function TMailNotifier.OpenDefaultMsgStore: HResult;
// this function finds the default message store.  It scans the MsgStoreTable for the
//  first entry with PR_DEFAULT_STORE set to true.  If none is found, the the default
//  store is assumed to be the first in the table
var
  iRow: integer;
  cbEntryID: ulong;
  EntryID: pEntryID;
  ObjType: ulong;
  // FMsgStoreTable: IMAPITable;
  hr: HResult;
  cRows: longint;
  pta: TSPropTagArray;
  iDefaultRow: longint;
  prs: PSRowSet;
  sc: longint;
  cb: integer;
  cValues: ulong;
  SPropValue: PSPropValue;
  FolderRoot: IMAPIFolder;
  Folder: IMAPIFolder;
  ProblemArray: PSPropProblemArray;
  RowsSought: longint;
  lpGUID: ^TGUID;

begin
  if not assigned(MAPISession)
    then raise exception.create('No MAPI session established');

  hr := MAPISession.GetMsgStoresTable(0, FMsgStoreTable);
  if failed(hr) then
    begin
      result := hr;
      exit;
    end;

  cRows := 0;
  hr := FMsgStoreTable.GetRowCount(0, cRows);

  if cRows = 0 then
    begin
      FMsgStoreTable._Release;
      result := MAPI_E_CALL_FAILED;
      exit;
    end;

  // look for the default message store row and return its entryID
  iDefaultRow := -1;

  for iRow := 0 to cRows - 1 do
    begin
      pta.cValues := 1;
      pta.aulPropTag[0] := PR_DEFAULT_STORE;

      // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter026_7x83.htm
      hr := FMsgStoreTable.SetColumns(@pta, 0);
      if failed(hr) then
      begin
        FMsgStoreTable._Release;
        result := hr;
        exit;
      end;

      // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter026_8xx3.htm
      RowsSought := 0;
      hr := FMsgStoreTable.SeekRow(BOOKMARK_BEGINNING, iRow, RowsSought);
      if failed(hr) then
      begin
        FMsgStoreTable._Release;
        result := hr;
        exit;
      end;
      // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter026_7wkz.htm
      hr := FMsgStoreTable.QueryRows(1, 0, prs);
      if failed(hr) then
      begin
        FMsgStoreTable._Release;
        result := hr;
        exit;
      end;

      if ( (prs.cRows = 1) AND (prs.aRow[0].lpProps[0].ulPropTag = PR_DEFAULT_STORE)
          AND (prs.aRow[0].lpProps[0].Value.b > 0 ) ) then
        begin
          iDefaultRow := iRow;
          MAPIFreeBuffer(prs);
          break;
        end;
      MAPIFreeBuffer(prs);
    end;  // for

  if iDefaultRow = -1 then
    begin
      FMsgStoreTable._Release;
      result := MAPI_E_CALL_FAILED;
      exit;
    end;

  pta.cValues := 1;
  pta.aulPropTag[0] := PR_ENTRYID;

  hr := FMsgStoreTable.SetColumns(@pta, 0);
  if failed(hr) then
    begin
      FMsgStoreTable._Release;
      result := hr;
      exit;
    end;

  RowsSought := 0;
  hr := FMsgStoreTable.SeekRow(BOOKMARK_BEGINNING, iDefaultRow, RowsSought);
  if failed(hr) then
    begin
      FMsgStoreTable._Release;
      result := hr;
      exit;
    end;

  hr := FMsgStoreTable.QueryRows(1, 0, prs);
  if failed(hr) then
    begin
      FMsgStoreTable._Release;
      result := hr;
      exit;
    end;

  FMsgStoreTable._Release;

  if ( (prs.cRows = 1) AND (prs.aRow[0].lpProps[0].ulPropTag = PR_ENTRYID) ) then
    begin
      cb := prs.aRow[0].lpProps[0].Value.bin.cb;
      sc := MAPIAllocateBuffer(cb, pointer(EntryID));
      if failed(sc) then
       begin
         MAPIFreeBuffer(prs);
         Result := sc;
         exit;
       end;

      pointer(EntryID) := prs.aRow[0].lpProps[0].Value.bin.lpb;
      cbEntryID := cb;
      result := hrSuccess;
    end;

    MAPIFreeBuffer(prs);

    // open the default message store...
    // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter023_59id.htm
    hr := MAPISession.OpenMsgStore(0, cbEntryID, EntryID, IMsgStore, MDB_WRITE, DefaultStore);
    if failed(hr) then
      begin
        result := hr;
        exit;
      end;


    // ensure there is an IPM (Inter-personal Messaging) subtree under this store
    // if one does not exist... then create one with the standard name...
    // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter022_44c3.htm
    pta.cValues := 1;
    pta.aulPropTag[0] := PR_IPM_SUBTREE_ENTRYID;
    hr := DefaultStore.GetProps(@pta, 0, cValues, SPropValue);
    if (failed(hr) or (not SPropValue.ulPropTag = PR_IPM_SUBTREE_ENTRYID)) then
      begin
        if succeeded(hr) then
          begin
            MAPIFreeBuffer(SPropValue);
            // open the root folder
            // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter031_3se1.htm
            lpGUID := nil;

            hr := DefaultStore.OpenEntry(0, PentryID(nil), lpGUID^, MAPI_MODIFY, ObjType, IUnknown(FolderRoot));
            if failed(hr) then
              begin
                DefaultStore._Release;
                DefaultStore := nil;
                result := hr;
                exit;
              end;

            // create a sub-folder called 'IPM_SUBTREE' under the root...
            // ref:
            lpGUID := nil;
            hr := FolderRoot.CreateFolder(FOLDER_GENERIC, 'IPM_SUBTREE', nil, lpGUID^,
                                          OPEN_IF_EXISTS, Folder);
            if failed(hr) then
              begin
                FolderRoot._Release;
                DefaultStore._Release;
                DefaultStore := nil;
                result := hr;
                exit;
              end;

            // release the root
            FolderRoot._Release;

            // save the new folder
            hr:= Folder.SaveChanges(KEEP_OPEN_READONLY);
            if failed(hr) then
              begin
                Folder._Release;
                DefaultStore._Release;
                DefaultStore := nil;
                result := hr;
                exit;
              end;

            // get the new folder's EntryID property and...
            pta.cValues := 1;
            pta.aulPropTag[0] := PR_ENTRYID;
            hr := Folder.GetProps(@pta, 0, cValues, SPropValue);
            if failed(hr) then
              begin
                Folder._Release;
                DefaultStore._Release;
                DefaultStore := nil;
                result := hr;
                exit;
              end;

            Folder._Release;

            if SPropValue.ulPropTag = PR_ENTRYID then
              begin
                SPropValue.ulPropTag := PR_IPM_SUBTREE_ENTRYID;
                ProblemArray := nil;
                hr := DefaultStore.SetProps(1, SPropValue, ProblemArray);
                if failed(hr) then
                  begin
                    MAPIFreeBuffer(SPropValue);
                    Folder._Release;
                    DefaultStore._Release;
                    DefaultStore := nil;
                    result := hr;
                    exit;
                  end;
              end;
            MAPIFreeBuffer(SPropValue);
          end;
      end
    else
      MAPIFreeBuffer(SPropValue);
  result := hr;     // get to here then fails ????
end;


function TMailNotifier.RegisterForNewMail: HResult;
var
  cbEntryID: ulong;
  EntryID: PEntryID;
  ExplicitClass: pchar;
  ulRowCount: ulong;
  ulObjType: ulong;
  hr: HResult;
  lpGUID: ^TGUID;

begin
  ulRowCount := 0;
  FNewMailNotify := false;

  // get the inbox folder object - EntryID will be set to id of inbox folder entry
  // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter031_7acy.htm
  hr := DefaultStore.GetReceiveFolder('IPM', 0, cbEntryID, EntryID, ExplicitClass);
  if ( (failed(hr)) or (not assigned(ExplicitClass)) or (ExplicitClass <> 'IPM') ) then
    begin
     if succeeded(hr) then
        begin
          MAPIFreeBuffer(EntryID);
          MAPIFreeBuffer(ExplicitClass);
        end;
      result := hr;
      exit;
  end;

  // get the root folder of the default message store
  lpGUID := nil;
  hr := DefaultStore.OpenEntry(0, nil, lpGUID^, (MAPI_MODIFY or MAPI_DEFERRED_ERRORS),
                               ulObjType, IUnknown(RootFolder));

  if failed(hr) then
    begin
      MAPIFreeBuffer(EntryID);
      MAPIFreeBuffer(ExplicitClass);
      result := hr;
      exit;
    end;

  // create the event sink and register for new mail notification
  FNewMailSink := TNewMailAdviseSink.Create;

  // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter031_43l1.htm
  hr := DefaultStore.Advise(cbEntryID, EntryID, (fnevNewMail or fnevObjectMoved), FNewMailSink, FNewMailConnection);
  MAPIFreeBuffer(EntryID);
  MAPIFreeBuffer(ExplicitClass);

  if failed(hr) then
    begin
      FNewMailSink.free;
      result := hr;
      exit;
    end;

  // according to the Microsoft KB example - we need to force an RPC call to
  // the server - apparently without which the notification will not work (??)
  // so the rest of this function is all in aid of forcing an RPC call...

  // get the hierarchy table for the root folder of the message store...
  // ref: http://premium.microsoft.com/msdn/library/sdkdoc/mapisdk/inter011_1sh1.htm
  hr := RootFolder.GetHierarchyTable(MAPI_DEFERRED_ERRORS, FRootHierarchyTable);
  if failed(hr) then
    begin
      FNewMailSink.free;
      result := hr;
      exit;
    end;

  hr := FRootHierarchyTable.GetRowCount(TBL_NOWAIT, ulRowCount);
  if failed(hr) then
    begin
      FNewMailSink.free;
      result := hr;
      exit;
    end;

  // if, by some miracle, we actually get this far successfully, then
  // registration is successful...

  FNewMailNotify := true;
  hr := HRSUCCESS;
end;


function TMailNotifier.UnregisterForNewMail: HResult;
begin
//  oops - should get around to finishing this one day...
//  do an unadvise and clean up stuff...   sheesh, I'm lazy!
end;

initialization
  TCOMObjectFactory.Create(ComServer, TNewMailAdviseSink,CLSID_NewMailAdviseSink,
                           'NewMailAdviseSink', 'New Mail Advise Sink PFS', ciMultiInstance);

end.
